// -----------------------------------------------------------------------
//   
//   Copyright 2003-2008 H. Peter Anvin - All Rights Reserved
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
//   Bostom MA 02111-1307, USA; either version 2 of the License, or
//   (at your option) any later version; incorporated herein by reference.
//
// -----------------------------------------------------------------------

//
// abc80.v
//
// Top-level module for ABC80 FPGA implementation
//
// Designed for Altera Nios development kit, Cyclone Edition with
// the Lancelot add-on board from www.fpga.nl
//

module abc80 (
	      input  	    clock_50,	// 50 MHz clock
	      input   [1:0] clock_24,	// 24 MHz clock (on two pins)
	      input   [1:0] clock_27,	// 27 MHz clock (on two pins)
	      input         ext_clock,	// External clock input

              inout         ps2_clk,	// PS/2 keyboard clock
              inout         ps2_dat,	// PS/2 keyboard data
	      
	      output  [9:0] ledr,	// Red LEDs
	      output  [7:0] ledg,	// Green LEDs
	      output  [6:0] s7_0,	// 7-segment LEDs
	      output  [6:0] s7_1,	// 7-segment LEDs
	      output  [6:0] s7_2,	// 7-segment LEDs
	      output  [6:0] s7_3,	// 7-segment LEDs
	      
	      input   [3:0] key_n,	// Pushbutton switches
	      input   [9:0] sw,		// Slide switches

	      output        fl_rst_n,   // Flash ROM RST#
	      output        fl_ce_n,    // Flash ROM CE#
	      output        fl_oe_n,	// Flash ROM OE#
	      output        fl_we_n,    // Flash ROM WE#
	      output [21:0] fl_a,       // Flash ROM address bus
	      inout   [7:0] fl_dq,      // Flash ROM data bus

	      output  [1:0] dram_ba,    // SDRAM bank selects
	      output        dram_ras_n,	// SDRAM RAS#
	      output        dram_cas_n,	// SDRAM CAS#
	      output        dram_cke,   // SDRAM clock enable
	      output        dram_clk,   // SDRAM clock
	      output        dram_cs_n,  // SDRAM CS#
	      output        dram_we_n,  // SDRAM WE#
	      output  [1:0] dram_dqm,   // SDRAM DQM (per byte)
	      output [11:0] dram_a,     // SDRAM address bus
	      inout  [15:0] dram_dq,    // SDRAM data bus
	      
	      output        sram_ce_n,	// SRAM CE#
	      output        sram_oe_n,	// SRAM OE#
	      output        sram_we_n,	// SRAM WE#
	      output  [1:0] sram_be_n,  // SRAM LB#, UB#
	      output [17:0] sram_a,	// SRAM address bus
	      inout  [15:0] sram_dq,    // SRAM data bus

	      output        sd_clk,     // SD card clock
	      inout         sd_cmd,     // SD card DI/MOSI/CMD
	      inout         sd_dat0,    // SD card SO/MISO/DAT0
	      inout         sd_dat3,    // SD card CS#/CD/DAT3
	      
	      output        uart_txd,   // RS232 port TxD
	      input         uart_rxd,   // RS232 port RxD
	      
	      output  [3:0] vga_r,	// VGA red
	      output  [3:0] vga_g,      // VGA green
	      output  [3:0] vga_b,	// VGA blue
	      output        vga_hs,	// VGA horz sync
	      output        vga_vs, 	// VGA vert sync

	      output        aud_xck,    // Audio master clock
	      output        aud_bclk,	// Audio bitclock
	      output        aud_dacdat, // Audio DAC data
	      output        aud_daclrck, // Audio DAC framing
	      input         aud_adcdat, // Audio ADC data
	      output        aud_adclrck, // Audio ADC framing
	      
	      inout  [35:0] gpio_0,   	// GPIO headers
	      inout  [35:0] gpio_1	// GPIO headers
	      );

   // ------------------------------------------------------------------------
   //  PLLs and clock distribution
   // ------------------------------------------------------------------------

   wire 	 cpu_clk;	//   25 MHz
   wire 	 video_clk;	//   25 MHz = screen pixel rate
   wire 	 fast_clk;	//  100 MHz
   wire 	 audio_clk;	//   16 MHz = I2S master clock

   assign 	 video_clk = cpu_clk;
   
   pll1 pll1 (
	      .inclk0 ( clock_50 ), // 50 MHz input clock
	      .c0 ( fast_clk ),	    // x2/1  = 100 MHz
	      .c1 ( cpu_clk ),      // x1/2  =  25 MHz
	      .c2 ( audio_clk )	    // x8/25 =  16 MHz
	      );
   
   // ------------------------------------------------------------------------
   //  Reset - synchronization and pulse stretch
   // ------------------------------------------------------------------------
`define RESET_WIDTH 6		// 2^6 = 64 cycles minimum width
   reg [`RESET_WIDTH:0] rst_ctr = 0;
   wire 		rst_n = rst_ctr[`RESET_WIDTH];
   reg 			power_rst_n = 1'b0;
   
   always @(posedge cpu_clk)
     if (~key_n[2])		// KEY2 = CPU reset
       rst_ctr <= 0;
     else if (~rst_n)
       rst_ctr <= rst_ctr + 1;

   always @(posedge cpu_clk)
     power_rst_n <= 1'b1;
   
   // ------------------------------------------------------------------------
   //  Standard LEDs (here so that we can switch around the LEDs more easily
   //  for quick debugging.)
   // ------------------------------------------------------------------------
   wire [7:0] 	 std_led;
   reg [6:0] 	 std_s7_1;
   reg [6:0] 	 std_s7_0;

   reg [7:0] 	 prog_led;
   reg [7:0] 	 prog_s7_1;
   reg [7:0] 	 prog_s7_0;
   reg [1:0] 	 prog_led_ctl;

   assign        ledr = sw;	// Show the switch status
   assign	 ledg = prog_led_ctl[0] ? prog_led : std_led;

   assign        s7_3 = ~prog_s7_1[6:0];
   assign        s7_2 = ~prog_s7_0[6:0];
   assign 	 s7_1 = ~std_s7_1[6:0];
   assign 	 s7_0 = ~std_s7_0[6:0];
   
   // ------------------------------------------------------------------------
   //  Timers
   // ------------------------------------------------------------------------

   // ABC80 has two interrupt-capable timers: one at 128 us (7812.5 Hz) and
   // one at 20 ms (50 Hz).  The 128 us clock was slightly off, since
   // the real ABC80 had 11.9808 MHz and not 12 MHz for its crystal
   // (the 50 Hz clock was exact), but everyone treated it as 128 us
   // (0.16% error).  We have a 50 MHz crystal; cpu_clk is already
   // divided by 2 to obtain 25 MHz (40 ns); we then divide it by 400 to
   // obtain a 16 us clock, which then is divided by 8 for 128 us and by
   // 625*2 for the 50 Hz clock.  Note the last stage is always divide by 2,
   // so we get 50% duty cycle.  I'm so anal-retentive.
   
   reg [8:0] 	 clk_div0;	// 16 us counter
   reg [2:0] 	 clk_div1;	// 128 us counter
   reg [9:0] 	 clk_div2;	// 10 ms counter
   reg 		 clk_div3;	// 20 ms counter
   
   wire		 ctr_128us;	// 128 us counter out
   wire		 ctr_20ms;	// 20 ms counter out

   always @(posedge cpu_clk)
     begin
	if ( clk_div0 == 0 )
	  clk_div0 <= 9'd399;
	else
	  clk_div0 <= clk_div0-1;
	  
	if ( clk_div0 == 0 )
	  begin
	     clk_div1 <= clk_div1 + 1; // No special wrapping needed
	     
	     if ( clk_div2 == 0 )
	       clk_div2 <= 10'd624;
	     else
	       clk_div2 <= clk_div2-1;

	     if ( clk_div2 == 0 )
	       clk_div3 <= ~clk_div3;
	  end // if ( clk_div0 == 0 )
     end // always @ (posedge cpu_clk)

   assign ctr_16us  = clk_div0[8];
   assign ctr_128us = clk_div1[2];
   assign ctr_20ms  = clk_div3;
	     
   // ------------------------------------------------------------------------
   //  Common bus wires declaration
   // ------------------------------------------------------------------------
   wire [15:0] 	 cpu_a;		// Output address from CPU (used for IORQ)
   wire [19:0] 	 mmu_a;		// Output address from MMU (within a device)
   reg  [7:0] 	 msel;		// Decoded device selects for MREQ
   reg  [7:0] 	 cpu_di;	// Input data to CPU
   wire [7:0] 	 cpu_do;	// Output data from CPU
   wire 	 cpu_m1_n;	// M1# from CPU
   wire 	 cpu_iorq_n;	// IORQ# from CPU
   wire 	 cpu_mreq_n;	// MREQ# from CPU
   wire 	 cpu_rd_n;	// RD# from CPU
   wire 	 cpu_wr_n;	// WR# from CPU
   wire 	 cpu_rfsh_n;	// RFSH# from CPU
   wire 	 cpu_halt_n;	// HALT# from CPU
   wire 	 cpu_busak_n;	// BUSAK# from CPU (not used)
   wire 	 cpu_int_n;	// INT# to CPU
   wire 	 cpu_nmi_n;	// NMI# to CPU
   // RETI# is a non-Z80 signal indicating that a RETI has been
   // executed.  We provide it in the CPU core because it is not feasible
   // for I/O devices to snoop the bus.
   wire 	 cpu_reti_n;	// RETI# from CPU
   
   // ABC-bus decoded strobes
   reg [7:0] 	 abc_out_n;
   reg [7:0] 	 abc_in_n;
   
   // ------------------------------------------------------------------------
   //  Keyboard controller
   // ------------------------------------------------------------------------

   wire [7:0] 	 kb_data;	// Keyboard symbol output
   wire [7:0] 	 kb_stat;	// Keyboard status word
   // Keyboard status word:
   // 7..0 = Strobe CapsLock NumLock ScrollLock Super Alt Ctrl Shift
   
   keyboard keyboard (
		      .clkin ( cpu_clk ),
		      .reset_n ( rst_n ),
		      .ps2_kclk ( ps2_clk ),
		      .ps2_kdata ( ps2_dat ),
		      .ps2_mclk ( ),
		      .ps2_mdata ( ),
		      .kb_data ( kb_data ),
		      .kb_stat ( kb_stat )
		      );

   // ------------------------------------------------------------------------
   //  Video controller
   // ------------------------------------------------------------------------

   wire [10:0] 	 video_a;
   wire [7:0] 	 video_d;
   wire [10:0] 	 chargen_a;
   wire [7:0] 	 chargen_d;
   wire [2:0] 	 rgb;
   wire [7:0] 	 vram_do;
   wire [7:0] 	 cgen_do;
   reg 		 video_width;
   
   videoram video_ram (
		       .wren_a ( 0 ),
		       .address_a ( video_a ),
		       .q_a ( video_d ),
		       .clock_a ( video_clk ),
		       .data_b ( cpu_do ),
		       .wren_b ( msel[2] & ~cpu_wr_n ),
		       .address_b ( mmu_a[11:0] ),
		       .q_b ( vram_do ),
		       .clock_b ( fast_clk )
		       );
   
   chargen char_gen (
		     .wren_a ( 0 ),
		     .address_a ( chargen_a ),
		     .q_a ( chargen_d ),
		     .clock_a ( video_clk ),
		     .data_b ( cpu_do ),
		     .wren_b ( msel[3] & ~cpu_wr_n ),
		     .address_b ( mmu_a[11:0] ),
		     .q_b ( cgen_do ),
		     .clock_b ( fast_clk )
		     );
      
   display video (
		  .clk ( video_clk ),
		  .width ( video_width ),
		  .reverse ( sw[7] ),
		  .noblink ( sw[6] ),
		  .testpattern ( sw[5] ),
		  .reveal ( ~key_n[0] ),
		  .a ( video_a ),
		  .d ( video_d ),
		  .ga ( chargen_a ),
		  .gd ( chargen_d ),
		  .rgb ( rgb ),
		  .vsync ( vga_vs ),
		  .hsync ( vga_hs )
		  );

   assign 	 vga_r = {4{rgb[2]}};
   assign 	 vga_g = {4{rgb[1]}};
   assign 	 vga_b = {4{rgb[0]}};

   // ------------------------------------------------------------------------
   //  External SRAM
   // ------------------------------------------------------------------------

   wire [7:0] 	  sram_do;		// Data out from sram
   wire		  sram_oe_w;
   wire		  sram_we_w;
   reg 		  sram_we_q1;
   reg 		  sram_we_q2;
   
   assign 	  sram_ce_n = ~msel[0];
   
   assign sram_oe_w = msel[0] & ~cpu_rd_n;
   assign sram_we_w = msel[0] & ~cpu_wr_n;

   always @(negedge rst_n or posedge fast_clk)
     if ( ~rst_n )
       begin
	  sram_we_q1 <= 0;
	  sram_we_q2 <= 0;
       end // if ( ~rst_n )
     else
       begin
	  // Delayed issues of WE# - to make sure the SRAM
	  // doesn't latch data late
	  sram_we_q1 <= sram_we_w;
	  sram_we_q2 <= sram_we_q1;
       end // else: !if( ~rst_n )

   // Driving output pins...
   assign sram_a    = mmu_a[18:1];
   assign sram_be_n = mmu_a[0] ? ~2'b10 : ~2'b01;
   
   assign sram_oe_n = ~sram_oe_w;
   assign sram_we_n = ~(sram_we_w & ~sram_we_q2);

   assign sram_dq = sram_we_w ? { cpu_do, cpu_do } : 16'bz;
   
   // SRAM Input side MUX
   assign sram_do = mmu_a[0] ? sram_dq[15:8] : sram_dq[7:0];

   // ------------------------------------------------------------------------
   //  External flash ROM
   //  Note: the flash ROM is a 70 ns part. With MMU delays, we have about
   //  25 ns in the initial cycle left at 25 MHz, so we need to delay for
   //  an additional two cycles (wait states).  In 3 and 6 MHz modes, those
   //  will vanish; in 12 and 25 MHz mode accessing the flash is thus slower
   //  than running out of RAM.  We're okay with that.
   //
   //  Also note the flash ROM is 4 MB in size.  Right now we can only access
   //  the low 1 MB due to MMU limitations.
   // ------------------------------------------------------------------------

   reg  [7:0] 	  flsh_do;	// Data out from flash
   wire 	  flsh_oe_w = msel[1] & ~cpu_rd_n;
   reg  [1:0] 	  flsh_wait_ctr = 2'b00;
   wire 	  flsh_wait_n = ~(flsh_oe_w & ~flsh_wait_ctr[1]);
   
   assign fl_rst_n = rst_n;
   assign fl_oe_n  = 1'b0;	// Flash always output
   assign fl_we_n  = 1'b1;	// Treat the flash as readonly
   assign fl_ce_n  = ~flsh_oe_w;
   assign fl_a     = { 2'b00, mmu_a };
   
   always @(negedge rst_n or posedge cpu_clk)
     if (~rst_n)
       flsh_wait_ctr <= 2'b00;
     else
       casex ({flsh_oe_w, flsh_wait_ctr})
	 3'b0xx: flsh_wait_ctr <= 2'b00;
	 3'b100: flsh_wait_ctr <= 2'b01;
	 3'b101: flsh_wait_ctr <= 2'b10;
	 3'b11x: flsh_wait_ctr <= 2'b10;
       endcase

   always @(posedge fast_clk)
     flsh_do <= fl_dq;
   
   // ------------------------------------------------------------------------
   //  BASIC patching ROM for 80-column mode
   // ------------------------------------------------------------------------

   wire [15:0] brompatch;

   // Patch rom: contains alternate contents for 80-column mode
   basic80 basic_patch (
			.address( mmu_a[7:0] ),
			.clock( fast_clk ),
			.q ( brompatch )
			);

   // The patching itself is now done in the CPU data input
   
   // ------------------------------------------------------------------------
   //  MMU and memory address decoding
   // ------------------------------------------------------------------------

   wire [15:0]   mmu_q;		// The actual bits coming out of the MMU
   wire [2:0] 	 mmu_devsel;	// Undecoded device selects
   wire 	 mmu_patch;	// Patch BASIC?
   
   reg [9:0] 	 mmu_mod_addr;	// Modify address holding register
   reg [7:0] 	 mmu_mod_data;	// Modify data holding register (low byte)

   wire 	 mmu_wr_e;	// MMU write enable
   wire [15:0] 	 mmu_rd_data;	// Read data output from MMU

   reg [5:0] 	 mmu_map_sel;	// XXWWRR
   wire [1:0] 	 mmu_map;

   assign 	 mmu_map =
		 ~cpu_m1_n ? mmu_map_sel[5:4] :
		 ~cpu_wr_n ? mmu_map_sel[3:2] :
		 mmu_map_sel[1:0];

   mmuram mmu (
	       .address_a ( { mmu_map, cpu_a[15:8] } ),
	       .data_a ( 16'bx ), // Never written
	       .q_a ( mmu_q ),
	       .wren_a ( 1'b0 ),
	       .address_b ( mmu_mod_addr[9:0] ),
	       .data_b ( { cpu_do, mmu_mod_data[7:0] } ),
	       .q_b ( mmu_rd_data ),
	       .wren_b ( mmu_wr_e ),
	       .clock ( fast_clk )
	       );

   assign 	 mmu_patch = mmu_q[15];
   assign 	 mmu_devsel = mmu_q[14:12];
   assign 	 mmu_a[19:8] = mmu_q[11:0];
   assign 	 mmu_a[7:0] = cpu_a[7:0];

   always @(posedge fast_clk)
     begin
	if ( ~cpu_mreq_n )
	  case ( mmu_devsel )
	    3'h0:   msel <= 8'b00000001;
	    3'h1:   msel <= 8'b00000010;
	    3'h2:   msel <= 8'b00000100;
	    3'h3:   msel <= 8'b00001000;
	    3'h4:   msel <= 8'b00010000;
	    3'h5:   msel <= 8'b00100000;
	    3'h6:   msel <= 8'b01000000;
	    3'h7:   msel <= 8'b10000000;
	  endcase // case( mmu_devsel )
	else
	  msel <= 8'b00000000;
     end

   // ------------------------------------------------------------------------
   //  PIO
   // ------------------------------------------------------------------------

   reg [7:0] pioa_ai;
   wire [7:0] pioa_do;
   wire [7:0] piob_do;
   wire [7:0] piob_ao;
   wire       pio_sel;
   wire       pioa_ieo;
   wire       piob_ieo;
   wire       pioa_int_n;
   wire       piob_int_n;

   wire       cas_relay;
   wire       cas_output;
   wire       cas_input;

   wire       v24_dtr;
   wire       v24_rts;
   wire       v24_cts;
   wire       v24_rxd;
   wire       v24_txd;

   wire       pr_txd;
   wire       pr_cts;
   
   always @(posedge cpu_clk)
     begin
	pioa_ai[6:0] <= kb_data[6:0]; // 7 bit ASCII
	pioa_ai[7]   <= kb_stat[7];   // Strobe
     end
   
   T80PIO pioa (
		.CLK_n ( cpu_clk ),
		.RESET_n ( rst_n ),
		.Di ( cpu_do ),
		.Do ( pioa_do ),
		.CE_n ( ~(pio_sel & ~cpu_a[1]) ),
		.CDsel ( cpu_a[0] ),
		.M1_n ( cpu_m1_n ),
		.IORQ_n ( cpu_iorq_n ),
		.RD_n ( cpu_rd_n ),
		.RETI_n ( cpu_reti_n ),
		.IEI ( 1 ),
		.IEO ( pioa_ieo ),
		.INT_n ( pioa_int_n ),
		.Ai ( pioa_ai ),
		.ASTB_n ( ctr_128us ),
		.BSTB_n ( /* == piob.astb_n */ 1 )
		);
   
   T80PIO piob (
		.CLK_n ( cpu_clk ),
		.RESET_n ( rst_n ),
		.Di ( cpu_do ),
		.Do ( piob_do ),
		.CE_n ( ~(pio_sel & cpu_a[1]) ),
		.CDsel ( cpu_a[0] ),
		.M1_n ( cpu_m1_n ),
		.IORQ_n ( cpu_iorq_n ),
		.RD_n ( cpu_rd_n ),
		.RETI_n ( cpu_reti_n ),
		.IEI ( pioa_ieo ),
		.IEO ( piob_ieo ),
		.INT_n ( piob_int_n ),
		.Ai ( { cas_input, 4'b1111, v24_dtr, v24_rts, v24_rxd } ),
		.Ao ( piob_ao ),
		.ASTB_n ( 1 ),
		.BSTB_n ( /* unused */ 1 )
		);

   // V.24 (RS232) port outputs
   assign     v24_txd = piob_ao[3];
   assign     v24_cts = piob_ao[4];

   // Pin assignments for V.24
   assign     v24_dtr = ~sw[3];
   assign     v24_rts = 1'b0;
   assign     v24_rxd = sw[3] ? uart_rxd : 1'b1;

   // Outbound UART port
   assign     uart_txd = sw[3] ? v24_txd : pr_txd;
   assign     pr_cts = 1'b0;
   
   // Cassette outputs
   assign     cas_relay  = piob_ao[5];
   assign     cas_output = piob_ao[6];

   // Map the cassette port to some of the very few unclaimed (by the
   // CF card) pins on Prototype Connector 1.
   //assign     p1_a29 = cas_relay;
   //assign     p1_a28 = cas_output;
   //assign     cas_input = p1_a38;
   
   assign     std_led[3] = cas_relay;
   assign     std_led[4] = cas_output;
   
   // ------------------------------------------------------------------------
   //  ABC-bus
   // ------------------------------------------------------------------------
   
   wire       abc_sel;
   
   always @(*)
     begin
	if ( abc_sel & ~cpu_wr_n )
	  case ( cpu_a[2:0] )
	    3'b000: abc_out_n = ~8'h01;
	    3'b001: abc_out_n = ~8'h02;
	    3'b010: abc_out_n = ~8'h04;
	    3'b011: abc_out_n = ~8'h08;
	    3'b100: abc_out_n = ~8'h10;
	    3'b101: abc_out_n = ~8'h20;
	    3'b110: abc_out_n = ~8'h40;
	    3'b111: abc_out_n = ~8'h80;
	  endcase // case( cpu_a[2:0] )
	else
	  abc_out_n = ~8'h00;
      
	if ( abc_sel & ~cpu_rd_n )
	  case ( cpu_a[2:0] )
	    3'b000: abc_in_n = ~8'h01;
	    3'b001: abc_in_n = ~8'h02;
	    3'b010: abc_in_n = ~8'h04;
	    3'b011: abc_in_n = ~8'h08;
	    3'b100: abc_in_n = ~8'h10;
	    3'b101: abc_in_n = ~8'h20;
	    3'b110: abc_in_n = ~8'h40;
	    3'b111: abc_in_n = ~8'h80;
	  endcase // case( cpu_a[2:0] )
	else
	  abc_in_n = ~8'h00;
     end // always @ (*)

   // SD card controller
   wire [7:0] abc_sd_di;
   wire       sd_select;
   wire       sd_active;

   sdcontroller sdcontroller (
			      .reset_n ( rst_n ),
			      .clk ( cpu_clk ),

			      .sd_cs_n ( sd_dat3 ),
			      .sd_di ( sd_cmd ),
			      .sd_clk ( sd_clk ),
			      .sd_do ( sd_dat0 ),

			      /* Terasic didn't hook up the Card Detect
			         and Write Enable switches in the SD socket
			         to FPGA pins, so we have to make do with
			         manual switches. */
			      .sd_cd_n ( ~sw[0] ), // SW0 on for inserted
			      .sd_we_n ( sw[1] ),  // SW1 on for Write Protect

			      .abc_do ( cpu_do ),
			      .abc_di ( abc_sd_di ),
			      .abc_out_n ( abc_out_n[0] ),
			      .abc_cs_n ( abc_out_n[1] ),
			      .abc_c1_n ( abc_out_n[2] ),
			      .abc_c2_n ( abc_out_n[3] ),
			      .abc_c3_n ( abc_out_n[4] ),
			      .abc_c4_n ( abc_out_n[5] ),
			      .abc_inp_n ( abc_in_n[0] ),
			      .abc_status_n ( abc_in_n[1] ),
			      .abc_rst_n ( abc_in_n[7] ),

			      .select ( sd_select ),
			      .active ( sd_active )
			      );

   // Printer controller
   wire [7:0] abc_pr_di;
   wire       pr_select;
   
   printer printer (
		    .reset_n ( rst_n ),
		    .clk ( cpu_clk ),
		    .tty_txd ( pr_txd ),
		    .tty_cts ( pr_cts ),
	            .abc_out_n ( abc_out_n[0] ),
		    .abc_cs_n ( abc_out_n[1] ),
		    .abc_c1_n ( abc_out_n[2] ),
		    .abc_c2_n ( abc_out_n[3] ),
		    .abc_c3_n ( abc_out_n[4] ),
		    .abc_inp_n ( abc_in_n[0] ),
		    .abc_status_n ( abc_in_n[1] ),
		    .abc_rst_n ( abc_in_n[7] ),
		    .abc_do ( cpu_do ),
		    .abc_di ( abc_pr_di ),
		    .select ( pr_select )
		    );

   // The terminology gets a bit funny there.  abc_do means data from
   // the ABC-bus to the main CPU.
   wire [7:0] abc_do = abc_sd_di & abc_pr_di;

   // Closest thing we get to a disk LED
   assign     std_led[0] = sd_active;
   assign     std_led[1] = sd_select;
   assign     std_led[2] = 1'b0; // Do something interesting here...
   
   // Video width control INP 4 = 80, INP 3 = 40.
   // Also allow KEY1 to flip it.
   wire       key1_debounced;
   wire       key1_strobe;

   debounce key1_debounce (
			   .clk ( cpu_clk ),
			   .reset_n ( rst_n ),
			   .in ( ~key_n[1] ),
			   .out ( key1_debounced ),
			   .strobe ( key1_strobe )
			   );

   always @(negedge rst_n or posedge cpu_clk)
     begin
	if ( ~rst_n )
	  video_width <= ~sw[9]; // Set SW9 to go to 40 column after reset
	else
	  if ( ~abc_in_n[3] )
	    video_width <= 1'b0;
	  else if ( ~abc_in_n[4] )
	    video_width <= 1'b1;
	  else if ( key1_strobe & key1_debounced )
	    video_width <= ~video_width;
     end

   // Sound generator
   reg [7:0] sound;
   always @(negedge rst_n or posedge cpu_clk)
     begin
	if ( ~rst_n )
	  sound <= 0;
	else
	  if ( ~abc_out_n[6] )
	    sound <= cpu_do;
     end

   // ------------------------------------------------------------------------
   //  Audio generation
   // ------------------------------------------------------------------------

   reg [7:0] sound_sc;
   
   always @(posedge audio_clk)
     sound_sc <= sound;		// Synchronize to audio_clk
   
   sound_i2s sound_i2s (
			.i2s_clk (audio_clk),
			
			.mixer_ctl ({sound_sc[5], sound_sc[3], sound_sc[4]}),
			.vco_sel (sound_sc[2]),
			.vco_pitch (sound_sc[1]),
			.envsel (sound_sc[7:6]),
			.inhibit (~sound_sc[0]),

			.i2s_dat (aud_dacdat),
			.i2s_lrck (aud_daclrck)
			);

   // Drive the output clocks via DDR buffers for better jitter performance
   ddio_out aud_xck_buf (
		     .datain_l ( 1'b0 ),
		     .datain_h ( 1'b1 ),
		     .oe ( 1'b1 ),
		     .outclock ( audio_clk ),
		     .dataout ( aud_xck )
		     );
   ddio_out aud_bclk_buf (
		      .datain_l ( 1'b0 ),
		      .datain_h ( 1'b1 ),
		      .oe ( 1'b1 ),
		      .outclock ( audio_clk ),
		      .dataout ( aud_bclk )
		      );

   assign    aud_adclrck = aud_daclrck;

   // ------------------------------------------------------------------------
   //  Internal non-ABC I/O registers (MMU, turbo, LED)
   // ------------------------------------------------------------------------

   wire       intio_sel; // Selected for access via I/O ports
   reg [7:0]  intio_do;	 // Data out from MMU I/O ports

   reg        turbo_set;
      
   assign mmu_wr_e = intio_sel & ~cpu_wr_n & (cpu_a[5:0] == 5'b00011);

   always @(negedge rst_n or posedge fast_clk)
     begin
	if ( !rst_n )
	  begin
	     mmu_mod_addr <= 0;
	     mmu_mod_data <= 0;
	     mmu_map_sel  <= 0;

	     turbo_set    <= 0;

	     prog_led_ctl <= 0;
	     prog_led	  <= 0;
	     prog_s7_0    <= 0;
	     prog_s7_1    <= 0;
	     
	     intio_do <= ~8'b0;
	  end
	else
	  begin
	     intio_do <= ~8'b0;
	     turbo_set <= 0;
	     
	     if ( intio_sel & ~cpu_wr_n )
	       begin
		  casex ( cpu_a[5:0] )
		    5'b00000:
		      mmu_mod_addr[7:0] <= cpu_do[7:0];
		    5'b00001:
		      mmu_mod_addr[9:8] <= cpu_do[1:0];
		    5'b00010:
		      mmu_mod_data <= cpu_do[7:0];
		    5'b001xx:
		      mmu_map_sel <= cpu_do[5:0];
		    5'b100xx:
		      turbo_set <= 1;
		    5'b10100:
		      prog_led_ctl <= cpu_do[1:0];
		    5'b10101:
		      prog_led <= cpu_do[7:0];
		    5'b10110:
		      prog_s7_0 <= cpu_do[7:0];
		    5'b10111:
		      prog_s7_1 <= cpu_do[7:0];
		  endcase // casex cpu_a[5:0] )
	       end // if ( intio_sel & ~cpu_wr_n )
	     else if ( ~abc_out_n[7] )
	       begin
		  // For compatibility with Mikrodatorn's 64K hack
		  mmu_map_sel <= { cpu_do[1:0], cpu_do[1:0], cpu_do[1:0] };
	       end
	     else if ( intio_sel & ~cpu_rd_n )
	       begin
		  casex ( cpu_a[5:0] )
		    5'b00000:
		      intio_do <= mmu_mod_addr[7:0];
		    5'b00001:
		      intio_do <= { 6'b0, mmu_mod_addr[9:8] };
		    5'b00010:
		      intio_do <= mmu_rd_data[7:0];
		    5'b00011:
		      intio_do <= mmu_rd_data[15:8];
		    5'b001xx:
		      intio_do <= { 2'b0, mmu_map_sel };
		    5'b100xx:
		      intio_do <= { 6'b0, cpu_turbo };
		    5'b10100:
		      intio_do <= { 6'b0, prog_led_ctl };
		    5'b10101:
		      intio_do <= prog_led;
		    5'b10110:
		      intio_do <= prog_s7_0;
		    5'b10111:
		      intio_do <= prog_s7_1;
		  endcase // casex ( cpu_a[5:0] )
	       end // if ( intio_sel & ~cpu_rd_n )
	  end
     end
   
   // Turbo control: KEY3 = toggle turbo
   wire key3_debounced;
   wire key3_strobe;
   reg [1:0] cpu_turbo;

   debounce key3_debounce (
			   .clk ( cpu_clk ),
			   .reset_n ( rst_n ),
			   .in ( ~key_n[3] ),
			   .out ( key3_debounced ),
			   .strobe ( key3_strobe )
			   );

   // This needs to operate in the cpu_clk domain, so use the
   // turbo_set register to communicate between decode and
   // this logic.  Note that we don't reset cpu_turbo on
   // a CPU reset.
   always @(posedge cpu_clk)
     if ( ~power_rst_n )
       cpu_turbo <= ~{2{sw[8]}}; // Set SW8 to go to 3 MHz mode on reset
     else if ( turbo_set )
       cpu_turbo <= cpu_do[1:0];
     else if ( key3_strobe & key3_debounced )
       cpu_turbo <= cpu_turbo - 1;
   
   // ------------------------------------------------------------------------
   //  I/O address decoding
   // ------------------------------------------------------------------------

   wire [7:0] io_do;
   
   // In original ABC80, A4 selected between the PIO and the bus controller
   // (which included the sound generator.)  We use A7 for our own devices.
   // The PIO wants to see IORQ# and M1# directly.
   
   assign     pio_sel = ~cpu_a[7] & cpu_a[4];
   assign     abc_sel = ~cpu_iorq_n & cpu_m1_n & ~cpu_a[7] & ~cpu_a[4];
   assign     intio_sel = ~cpu_iorq_n & cpu_m1_n & cpu_a[7];  // Internal ports
   
   // All I/O devices emit all ones when not selected, so we just AND together
   // the outputs.
   assign     io_do = pioa_do & piob_do & intio_do & abc_do;

   // ------------------------------------------------------------------------
   //  Main CPU
   // ------------------------------------------------------------------------

   assign     cpu_int_n = pioa_int_n & piob_int_n;
   assign     cpu_nmi_n = ctr_20ms;
   wire       cpu_wait_n = 1;	// Nothing to wait for atm
   wire       cpu_clk_en;

   always @(*)
     begin
	if ( ~cpu_mreq_n & mmu_patch & video_width &
	     ( brompatch[15:8] == mmu_a[15:8] ) )
	  cpu_di <= brompatch[7:0];
	else
	  case ( msel )
	    8'b00000001: cpu_di <= sram_do;
	    8'b00000010: cpu_di <= flsh_do;
	    8'b00000100: cpu_di <= vram_do;
	    8'b00001000: cpu_di <= cgen_do;
	    8'b10000000: cpu_di <= mmu_a[15:8];
	    8'b00000000: cpu_di <= io_do;
	    default:     cpu_di <= 8'bx;
	  endcase // case( msel )
     end // always @ (*)

   // If cpu_turbo is 0, we want to execute at 3 MHz instead of 25
   // ... which means enabling the clock 3 times every 25 cycles
   //
   //        0               1     
   // cycle  0123456789ABCDEF012345678
   //  3 MHz -------X-------X-------X-
   //  6 MHz ---X---X---X---X---X---X-
   // 12 MHz -X-X-X-X-X-X-X-X-X-X-X-X-
   // 25 MHz XXXXXXXXXXXXXXXXXXXXXXXXX
   //
   reg [4:0] cpu_slow_ctr;
   always @(posedge cpu_clk)
     begin
	if ( &cpu_slow_ctr[4:3] ) // count 00h..18h
	  cpu_slow_ctr <= 5'd0;
	else
	  cpu_slow_ctr <= cpu_slow_ctr + 1;
     end

   reg [2:0] cpu_slow_mask;

   always @(posedge cpu_clk)
     begin
	case ( cpu_turbo )
	  2'b00:
	    cpu_slow_mask <= 3'b000;
	  2'b01:
	    cpu_slow_mask <= 3'b100;
	  2'b10:
	    cpu_slow_mask <= 3'b110;
	  2'b11:
	    cpu_slow_mask <= 3'b111;
	endcase // case( cpu_turbo )
     end // always @ (posedge cpu_clk)
   
   assign cpu_clk_en = &(cpu_slow_ctr[2:0] | cpu_slow_mask) & flsh_wait_n;

   assign std_led[7:5] = cpu_slow_mask;
   
   always @(*)
     case ( cpu_turbo )
       2'b00:
	 begin
	    std_s7_1 <= 7'b0000000; // blank
	    std_s7_0 <= 7'b1001111; // 3
	 end
       2'b01:
	 begin
	    std_s7_1 <= 7'b0000000; // blank
	    std_s7_0 <= 7'b1111101; // 6
	 end
       2'b10:
	 begin
	    std_s7_1 <= 7'b0000110; // 1
	    std_s7_0 <= 7'b1011011; // 2
	 end
       2'b11:
	 begin
	    std_s7_1 <= 7'b1011011; // 2
	    std_s7_0 <= 7'b1101101; // 5
	 end
     endcase // case( cpu_turbo )

   T80se cpu (
	      .clk_n ( cpu_clk ),
	      .clken ( cpu_clk_en ),
	      .reset_n ( rst_n ),
	      .wait_n ( cpu_wait_n ),
	      .int_n ( cpu_int_n ),
	      .nmi_n ( cpu_nmi_n ),
	      .busrq_n ( 1 ),
	      .busak_n ( cpu_busak_n ),
	      .reti_n ( cpu_reti_n ),
	      .m1_n ( cpu_m1_n ),
              .mreq_n ( cpu_mreq_n ),
              .iorq_n ( cpu_iorq_n ),
              .rd_n ( cpu_rd_n ),
              .wr_n ( cpu_wr_n ),
              .rfsh_n ( cpu_rfsh_n ),
              .halt_n ( cpu_halt_n ),
              .a ( cpu_a ),
              .di ( cpu_di ),
              .do ( cpu_do )
	      );
   
endmodule // abc80
